"""Test the AirPatrol API client."""

from typing import Any
from unittest.mock import AsyncMock, MagicMock

import pytest
from aiohttp import ClientSession

from airpatrol.api import (
    AirPatrolAPI,
    AirPatrolAuthenticationError,
    AirPatrolError,
)


@pytest.fixture
def mock_session() -> MagicMock:
    """Mock aiohttp session."""
    return MagicMock(spec=ClientSession)


@pytest.fixture
def api(mock_session: MagicMock) -> AirPatrolAPI:
    """Create AirPatrol API instance."""
    return AirPatrolAPI(mock_session, "test_access_token", "test_user_id")


@pytest.mark.asyncio
async def test_authenticate_success(mock_session: MagicMock) -> None:
    """Test successful authentication."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {
        "status": "ok",
        "entities": {
            "users": {
                "list": [
                    {
                        "id": "test_user_id",
                    }
                ]
            }
        },
        "misc": {"accessToken": "test_access_token"},
    }
    mock_session.post.return_value.__aenter__.return_value = mock_response

    api = await AirPatrolAPI.authenticate(mock_session, "test@example.com", "password")

    assert api._access_token == "test_access_token"
    assert api._user_id == "test_user_id"


@pytest.mark.asyncio
async def test_authenticate_failure_status(mock_session: MagicMock) -> None:
    """Test authentication failure with non-200 status."""
    mock_response = AsyncMock()
    mock_response.status = 401
    mock_response.text.return_value = "Unauthorized"
    mock_session.post.return_value.__aenter__.return_value = mock_response

    with pytest.raises(AirPatrolAuthenticationError, match="Authentication failed"):
        await AirPatrolAPI.authenticate(mock_session, "test@example.com", "password")


@pytest.mark.asyncio
async def test_authenticate_failure_exception(mock_session: MagicMock) -> None:
    """Test authentication failure with exception."""
    mock_session.post.side_effect = Exception("Connection error")

    with pytest.raises(Exception, match="Connection error"):
        await AirPatrolAPI.authenticate(mock_session, "test@example.com", "password")


@pytest.mark.asyncio
async def test_get_unique_id_authenticated(api: AirPatrolAPI) -> None:
    """Test getting unique ID when authenticated."""
    assert api.get_unique_id() == "test_user_id"


@pytest.mark.asyncio
async def test_get_access_token_authenticated(api: AirPatrolAPI) -> None:
    """Test getting access token after authentication."""
    assert api.get_access_token() == "test_access_token"


@pytest.mark.asyncio
async def test_get_pairings_success(api: AirPatrolAPI, mock_session: MagicMock) -> None:
    """Test successful pairings fetch."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {
        "status": "ok",
        "entities": {
            "pairingUser": {
                "list": [
                    {"userId": "test_user_id", "pairingId": "00000", "id": "000011"}
                ]
            },
            "pairings": {
                "list": [
                    {
                        "id": "00000",
                        "appId": "00000-00000-00000-00000-00000",
                        "cid": "00000",
                        "hwid": "AAAAA-BBBBB-CCCCC-DDDDD-EEEEE",
                        "name": "Test Device",
                        "type": "apw",
                    }
                ]
            },
        },
        "misc": {},
        "errors": [],
    }
    mock_session.get.return_value.__aenter__.return_value = mock_response

    data = await api.get_pairings()

    assert data["status"] == "ok"
    assert "entities" in data
    assert "pairingUser" in data["entities"]
    assert "pairings" in data["entities"]


@pytest.mark.asyncio
async def test_get_pairings_failure_status(
    api: AirPatrolAPI, mock_session: MagicMock
) -> None:
    """Test pairings fetch failure with non-200 status."""
    mock_response = AsyncMock()
    mock_response.status = 401
    mock_response.text.return_value = "Unauthorized"
    mock_session.get.return_value.__aenter__.return_value = mock_response

    with pytest.raises(AirPatrolAuthenticationError, match="Failed to fetch pairings"):
        await api.get_pairings()


@pytest.mark.asyncio
async def test_get_pairings_failure_error_status(
    api: AirPatrolAPI, mock_session: MagicMock
) -> None:
    """Test pairings fetch failure with error status in response."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {"status": "error", "errors": ["Invalid token"]}
    mock_session.get.return_value.__aenter__.return_value = mock_response
    with pytest.raises(AirPatrolError, match="API returned error status"):
        await api.get_pairings()


@pytest.mark.asyncio
async def test_get_devices_success(api: AirPatrolAPI, mock_session: MagicMock) -> None:
    """Test successful devices fetch."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {
        "status": "ok",
        "entities": {
            "pairingUser": {
                "list": [
                    {"userId": "test_user_id", "pairingId": "00000", "id": "000011"}
                ]
            },
            "pairings": {
                "list": [
                    {
                        "id": "00000",
                        "appId": "00000-00000-00000-00000-00000",
                        "cid": "00000",
                        "hwid": "AAAAA-BBBBB-CCCCC-DDDDD-EEEEE",
                        "name": "Test Device",
                        "type": "apw",
                    }
                ]
            },
        },
        "misc": {},
        "errors": [],
    }
    mock_session.get.return_value.__aenter__.return_value = mock_response

    devices = await api.get_devices()

    assert len(devices) == 1
    device = devices[0]
    assert device["id"] == "00000"
    assert device["name"] == "Test Device"
    assert device["type"] == "apw"
    assert device["hwid"] == "AAAAA-BBBBB-CCCCC-DDDDD-EEEEE"


@pytest.mark.asyncio
async def test_get_devices_no_user_pairings(
    api: AirPatrolAPI, mock_session: MagicMock
) -> None:
    """Test devices fetch when user has no pairings."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {
        "status": "ok",
        "entities": {
            "pairingUser": {
                "list": [
                    {
                        "userId": "other_user_id",  # Different user
                        "pairingId": "00000",
                        "id": "000011",
                    }
                ]
            },
            "pairings": {
                "list": [
                    {
                        "id": "00000",
                        "appId": "00000-00000-00000-00000-00000",
                        "cid": "00000",
                        "hwid": "AAAAA-BBBBB-CCCCC-DDDDD-EEEEE",
                        "name": "Vardagsrum",
                        "type": "apw",
                    }
                ]
            },
        },
        "misc": {},
        "errors": [],
    }
    mock_session.get.return_value.__aenter__.return_value = mock_response

    devices = await api.get_devices()

    assert len(devices) == 0


# Helper for aiohttp context manager mock
class MockAiohttpResponse:
    """Mock aiohttp response."""

    def __init__(
        self, status: int, json_data: Any | None = None, text_data: Any | None = None
    ) -> None:
        """Initialize mock aiohttp response."""
        self.status = status
        self._json = json_data
        self._text = text_data or ""
        self.url = "mock://url"

    async def __aenter__(self) -> "MockAiohttpResponse":
        """Enter aiohttp context manager."""
        return self

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> bool:
        """Exit aiohttp context manager."""
        return False

    async def json(self) -> Any:
        """Return JSON data."""
        return self._json

    async def text(self) -> Any:
        """Return text data."""
        return self._text


@pytest.mark.asyncio
async def test_get_data_success(api: AirPatrolAPI, mock_session: MagicMock) -> None:
    """Test successful data fetch."""

    def _mock_get(url: str, *args: Any, **kwargs: Any) -> MockAiohttpResponse:
        if "pairings" in url:
            return MockAiohttpResponse(
                200,
                json_data={
                    "entities": {
                        "pairingUser": {
                            "list": [
                                {
                                    "userId": "11111",
                                    "pairingId": "00000",
                                    "id": "000011",
                                }
                            ]
                        },
                        "users": {
                            "list": [
                                {"id": "11111", "firstName": None, "lastName": None}
                            ]
                        },
                        "pairings": {
                            "list": [
                                {
                                    "id": "00000",
                                    "appId": "00000-00000-00000-00000-00000",
                                    "cid": "00000",
                                    "hwid": "AAAAA-BBBBB-CCCCC-DDDDD-EEEEE",
                                    "name": "Test Device",
                                    "type": "apw",
                                }
                            ]
                        },
                    },
                    "misc": {},
                    "status": "ok",
                    "errors": [],
                },
            )
        if "command" in url:
            return MockAiohttpResponse(
                200,
                json_data={
                    "ApiVersion": "12",
                    "CommandMode": "parameters",
                    "ParametersData": {
                        "PumpPower": "on",
                        "PumpTemp": "16.000",
                        "PumpMode": "heat",
                        "FanSpeed": "max",
                        "Swing": "off",
                    },
                    "RoomTemp": "27.517",
                    "RoomHumidity": "56",
                },
            )
        return MockAiohttpResponse(404, json_data={})

    mock_session.get.side_effect = _mock_get
    api._access_token = "dummy_token"
    api._user_id = "11111"
    data = await api.get_data()
    assert isinstance(data, list)
    assert len(data) == 1
    unit = data[0]
    assert unit["unit_id"] == "00000"
    assert unit["name"] == "Test Device"
    assert unit["model"] == "AirPatrol APW"


@pytest.mark.asyncio
async def test_get_data_failure(api: AirPatrolAPI, mock_session: MagicMock) -> None:
    """Test data fetch failure due to non-auth error."""

    def _mock_get(url: str, *args: Any, **kwargs: Any) -> MockAiohttpResponse:
        if "pairings" in url:
            return MockAiohttpResponse(500, text_data="Internal Server Error")
        return MockAiohttpResponse(404, json_data={})

    mock_session.get.side_effect = _mock_get
    api._access_token = "dummy_token"
    api._user_id = "197090"
    with pytest.raises(AirPatrolError):
        await api.get_data()


@pytest.mark.asyncio
async def test_get_units(api: AirPatrolAPI, mock_session: MagicMock) -> None:
    """Test getting units from AirPatrol."""

    def _mock_get(url: str, *args: Any, **kwargs: Any) -> MockAiohttpResponse:
        if "pairings" in url:
            return MockAiohttpResponse(
                200,
                json_data={
                    "entities": {
                        "pairingUser": {
                            "list": [
                                {
                                    "userId": "11111",
                                    "pairingId": "00000",
                                    "id": "000011",
                                }
                            ]
                        },
                        "users": {
                            "list": [
                                {"id": "11111", "firstName": None, "lastName": None}
                            ]
                        },
                        "pairings": {
                            "list": [
                                {
                                    "id": "00000",
                                    "appId": "00000-00000-00000-00000-00000",
                                    "cid": "00000",
                                    "hwid": "AAAAA-BBBBB-CCCCC-DDDDD-EEEEE",
                                    "name": "Test Device",
                                    "type": "apw",
                                }
                            ]
                        },
                    },
                    "misc": {},
                    "status": "ok",
                    "errors": [],
                },
            )
        if "command" in url:
            return MockAiohttpResponse(
                200,
                json_data={
                    "ApiVersion": "12",
                    "CommandMode": "parameters",
                    "ParametersData": {
                        "PumpPower": "on",
                        "PumpTemp": "16.000",
                        "PumpMode": "heat",
                        "FanSpeed": "max",
                        "Swing": "off",
                    },
                    "RoomTemp": "27.517",
                    "RoomHumidity": "56",
                },
            )
        return MockAiohttpResponse(404, json_data={})

    mock_session.get.side_effect = _mock_get
    api._access_token = "dummy_token"
    api._user_id = "11111"
    units = await api.get_units()
    assert isinstance(units, list)
    assert len(units) == 1
    unit = units[0]
    assert unit["unit_id"] == "00000"
    assert unit["name"] == "Test Device"
    assert unit["model"] == "AirPatrol APW"


@pytest.mark.asyncio
async def test_pairings_cache(api: AirPatrolAPI, mock_session: MagicMock) -> None:
    """Test that pairings are cached after first fetch."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {
        "status": "ok",
        "entities": {
            "pairingUser": {
                "list": [
                    {"userId": "test_user_id", "pairingId": "00000", "id": "000011"}
                ]
            },
            "pairings": {
                "list": [
                    {
                        "id": "00000",
                        "appId": "00000-00000-00000-00000-00000",
                        "cid": "00000",
                        "hwid": "AAAAA-BBBBB-CCCCC-DDDDD-EEEEE",
                        "name": "Vardagsrum",
                        "type": "apw",
                    }
                ]
            },
        },
        "misc": {},
        "errors": [],
    }
    mock_session.get.return_value.__aenter__.return_value = mock_response

    # First call should fetch from API
    await api.get_pairings()
    mock_session.get.assert_called_once()

    # Second call should use cache
    await api.get_pairings()
    # Should still only be called once (no additional API calls)
    mock_session.get.assert_called_once()


@pytest.mark.asyncio
async def test_get_unit_climate_data(
    api: AirPatrolAPI, mock_session: MagicMock
) -> None:
    """Test getting climate data for a unit."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {
        "ApiVersion": "12",
        "CommandMode": "parameters",
        "ParametersData": {
            "PumpPower": "on",
            "PumpTemp": "22.000",
            "PumpMode": "heat",
            "FanSpeed": "max",
            "Swing": "off",
        },
        "RoomTemp": "22.5",
        "RoomHumidity": "45",
    }
    mock_session.get.return_value.__aenter__.return_value = mock_response

    result = await api.get_unit_climate_data("00000")

    # Verify the correct endpoint and headers were used
    mock_session.get.assert_called_once()
    call_args = mock_session.get.call_args
    assert "https://api.apsrvd.io/12/command" in str(call_args[0][0])
    assert call_args[1]["headers"]["Authorization"] == "Bearer test_access_token"
    assert call_args[1]["headers"]["X-Pairing-Id"] == "00000"

    assert result["RoomTemp"] == "22.5"
    assert result["ParametersData"]["PumpPower"] == "on"


@pytest.mark.asyncio
async def test_set_unit_climate_data(
    api: AirPatrolAPI, mock_session: MagicMock
) -> None:
    """Test setting climate data for a unit."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.json.return_value = {
        "ApiVersion": "12",
        "CommandMode": "parameters",
        "ParametersData": {
            "PumpPower": "on",
            "PumpTemp": "25.000",
            "PumpMode": "cool",
            "FanSpeed": "med",
            "Swing": "on",
        },
        "RoomTemp": "24.5",
        "RoomHumidity": "50",
    }
    mock_session.post.return_value.__aenter__.return_value = mock_response

    climate_data = {
        "ApiVersion": "12",
        "CommandMode": "parameters",
        "ParametersData": {
            "PumpPower": "on",
            "PumpTemp": "25.000",
            "PumpMode": "cool",
            "FanSpeed": "med",
            "Swing": "on",
        },
    }

    result = await api.set_unit_climate_data("00000", climate_data)

    # Verify the correct endpoint and headers were used
    mock_session.post.assert_called_once()
    call_args = mock_session.post.call_args
    assert "https://api.apsrvd.io/12/command" in str(call_args[0][0])
    assert call_args[1]["headers"]["Authorization"] == "Bearer test_access_token"
    assert call_args[1]["headers"]["X-Pairing-Id"] == "00000"
    assert call_args[1]["json"] == climate_data

    # Verify the response data is returned
    assert result["RoomTemp"] == "24.5"
    assert result["ParametersData"]["PumpPower"] == "on"


@pytest.mark.asyncio
async def test_set_unit_climate_data_failure(
    api: AirPatrolAPI, mock_session: MagicMock
) -> None:
    """Test set_unit_climate_data returns error."""

    def _mock_post(url: str, *args: Any, **kwargs: Any) -> MockAiohttpResponse:
        return MockAiohttpResponse(400, text_data="Bad Request")

    mock_session.post.side_effect = _mock_post
    api._access_token = "dummy_token"
    api._user_id = "11111"
    with pytest.raises(AirPatrolError):
        await api.set_unit_climate_data("00000", {"foo": "bar"})


@pytest.mark.asyncio
async def test_get_unit_climate_data_failure(
    api: AirPatrolAPI, mock_session: MagicMock
) -> None:
    """Test get_unit_climate_data returns error."""

    def _mock_get(url: str, *args: Any, **kwargs: Any) -> MockAiohttpResponse:
        return MockAiohttpResponse(400, text_data="Bad Request")

    mock_session.get.side_effect = _mock_get
    api._access_token = "dummy_token"
    api._user_id = "11111"
    with pytest.raises(AirPatrolError):
        await api.get_unit_climate_data("00000")



"""AirPatrol API client."""

from __future__ import annotations

import logging
from typing import Any, Dict, cast

from aiohttp import ClientSession

from .const import (
    API_BASE_URL,
    API_COMMAND_ENDPOINT,
    AUTH_BASE_URL,
    AUTH_LOGIN_ENDPOINT,
    AUTH_PAIRINGS_ENDPOINT,
)

_LOGGER = logging.getLogger(__name__)


class AirPatrolAuthenticationError(Exception):
    """Exception raised when authentication fails."""


class AirPatrolError(Exception):
    """Exception raised when an unexpected error occurs."""


class AirPatrolAPI:
    """AirPatrol API client."""

    def __init__(
        self, session: ClientSession, access_token: str, user_id: str | None = None
    ) -> None:
        """Initialize."""
        self._session = session
        self._access_token = access_token
        self._user_id = user_id
        self._pairings_cache: dict[str, Any] | None = None

    @staticmethod
    async def authenticate(
        session: ClientSession, email: str, password: str
    ) -> AirPatrolAPI:
        """Test if we can authenticate with the host.

        Returns
        -------
        AirPatrolAPI instance if authentication is successful

        """
        async with session.post(
            f"{AUTH_BASE_URL}{AUTH_LOGIN_ENDPOINT}",
            json={"email": email, "password": password},
        ) as resp:
            _LOGGER.debug("Authenticating with %s", resp.url)
            if resp.status != 200:
                _LOGGER.error("Authentication failed with status %s", resp.status)
                _LOGGER.error("Response: %s", await resp.text())
                raise AirPatrolAuthenticationError("Authentication failed")

            data = await resp.json()
            if data.get("status") == "ok":
                user_id = data["entities"]["users"]["list"][0]["id"]
                access_token = data["misc"]["accessToken"]
                return AirPatrolAPI(session, access_token, user_id)
        raise AirPatrolAuthenticationError("Authentication failed")

    def get_unique_id(self) -> str:
        """Get unique ID for the user."""
        if self._user_id is None:
            raise ValueError("Not authenticated")
        return self._user_id

    def get_access_token(self) -> str | None:
        """Get the access token."""
        return self._access_token

    def clear_pairings_cache(self) -> None:
        """Clear the pairings cache to force a fresh fetch."""
        self._pairings_cache = None

    async def get_pairings(self) -> dict[str, Any]:
        """Get pairings (devices) from AirPatrol API."""
        # Return cached data if available
        if self._pairings_cache is not None:
            _LOGGER.debug("Returning cached pairings data")
            return self._pairings_cache

        headers = {"Authorization": f"Bearer {self._access_token}"}
        async with self._session.get(
            f"{AUTH_BASE_URL}{AUTH_PAIRINGS_ENDPOINT}",
            headers=headers,
        ) as resp:
            _LOGGER.debug("Fetching pairings from %s", resp.url)
            if resp.status in [401, 403]:
                _LOGGER.error("Authentication failed with status %s", resp.status)
                _LOGGER.error("Response: %s", await resp.text())
                raise AirPatrolAuthenticationError("Failed to fetch pairings")

            if resp.status == 200:
                data = cast(Dict[str, Any], await resp.json())
                if data.get("status") != "ok":
                    _LOGGER.error("API returned error status: %s", data.get("errors"))
                    raise AirPatrolError("API returned error status")

                # Cache the successful response
                self._pairings_cache = data
                _LOGGER.debug("Cached pairings data")
                return data

        raise AirPatrolError(f"Failed to fetch pairings: {resp.status}")

    async def get_devices(self) -> list[dict[str, Any]]:
        """Get list of devices (pairings) for the authenticated user."""
        data = await self.get_pairings()

        # Get the user's pairings
        user_pairings = []
        pairing_users = data.get("entities", {}).get("pairingUser", {}).get("list", [])

        for pairing_user in pairing_users:
            if pairing_user.get("userId") == self._user_id:
                pairing_id = pairing_user.get("pairingId")
                # Find the corresponding pairing details
                pairings = data.get("entities", {}).get("pairings", {}).get("list", [])
                for pairing in pairings:
                    if pairing.get("id") == pairing_id:
                        user_pairings.append(pairing)
                        break

        return user_pairings

    async def get_data(self) -> list[dict[str, Any]]:
        """Get data from AirPatrol."""
        devices = await self.get_devices()
        units = []
        for device in devices:
            unit_id = device.get("id")
            if not isinstance(unit_id, str):
                continue  # Skip if unit_id is not a string
            # Get climate data for this unit
            climate_data = await self.get_unit_climate_data(unit_id)

            unit_data = {
                "unit_id": unit_id,
                "name": device.get("name", f"AirPatrol Device {unit_id}"),
                "model": f"AirPatrol {device.get('type', 'Unknown').upper()}",
                "manufacturer": "AirPatrol",
                "hwid": device.get("hwid"),
                "type": device.get("type"),
                "app_id": device.get("appId"),
                # Climate data
                "climate": climate_data,
            }
            units.append(unit_data)
        return units

    async def get_units(self) -> list[dict[str, Any]]:
        """Get list of units from AirPatrol."""
        return await self.get_data()

    async def get_unit_climate_data(self, unit_id: str) -> dict[str, Any]:
        """Get climate data for a specific unit."""
        headers = {
            "Authorization": f"Bearer {self._access_token}",
            "X-Pairing-Id": unit_id,
        }
        async with self._session.get(
            f"{API_BASE_URL}{API_COMMAND_ENDPOINT}",
            headers=headers,
        ) as resp:
            _LOGGER.debug(
                "Fetching climate data for unit %s from %s", unit_id, resp.url
            )
            if resp.status in [401, 403]:
                _LOGGER.error(
                    "Failed to fetch climate data for unit %s with status %s",
                    unit_id,
                    resp.status,
                )
                raise AirPatrolAuthenticationError("Failed to fetch climate data")

            if resp.status == 200:
                return cast(Dict[str, Any], await resp.json())

        raise AirPatrolError(f"Failed to fetch climate data: {resp.status}")

    async def set_unit_climate_data(
        self, unit_id: str, climate_data: dict[str, Any]
    ) -> dict[str, Any]:
        """Set climate data for a specific unit."""
        headers = {
            "Authorization": f"Bearer {self._access_token}",
            "X-Pairing-Id": unit_id,
        }
        async with self._session.post(
            f"{API_BASE_URL}{API_COMMAND_ENDPOINT}",
            headers=headers,
            json=climate_data,
        ) as resp:
            _LOGGER.debug("Setting climate data for unit %s", unit_id)
            if resp.status in [401, 403]:
                _LOGGER.error(
                    "Failed to set climate data for unit %s with status %s",
                    unit_id,
                    resp.status,
                )
                raise AirPatrolAuthenticationError("Failed to set climate data")

            if resp.status == 200:
                return cast(Dict[str, Any], await resp.json())

        raise AirPatrolError(f"Failed to set climate data: {resp.status}")



"""AirPatrol constants."""

AUTH_BASE_URL = "https://auth.apsrvd.io/v1"
AUTH_LOGIN_ENDPOINT = "/login"
AUTH_PAIRINGS_ENDPOINT = "/pairings"

API_VERSION = "12"
API_BASE_URL = f"https://api.apsrvd.io/{API_VERSION}"
API_COMMAND_ENDPOINT = "/command"
